c     07nov28  David Rupke  Added input guess for continuum.  Anything
c                           less than or equal to zero defaults to 
c                           the in-program algorithm.

c     6/27/94 - added wguess so you can send in a guess for the central
c     wavelength (to avoid getting munged by noise peaks).  Similar to
c     the absorption-line voigt routine.  Sending in 0 for wguess lets
c     it choose itself.

c     1/9/94 - Substituting Ted's new voi() bumped the intensities up by
c     sqrt(2pi), and the initial guess for intensity in evoigt needed to
c     be changed to make up for that.

c     11/6/93 - substituted Ted's new version of voi() which gives
c     correct intensity.  Therefore any program of mine older than
c     11/6/93 should be recompiled. - BW

c     7/9/92 - modifications made to allow a slightly more graceful
c     failure out of gaussj_2.  flag is returned as zero if the fit is
c     OK and 1 if something bad happens

      subroutine evoigt (wave,flux,sigma,num,sg,sl,mfit,fit,err, chisq
     $     ,flag,guess)

c     'mfit' parameter: Allowed values are 1-5, and indicate the
c     parameters that are fit.  If a parameter is not fit, it is fixed
c     at the initial guess (?).
c
c     1 = continuum
c     2 = cont. + intensity
c     3 = cont. + int. + central wavelength
c     4 = cont. + int. + central w.l. + Gaussian width
c     5 = cont. + int. + central w.l. + G. width + Lorentzian width


c  This program fits a Voigt emission line profile to a spectrum.

      real flux(num),wave(num),sigma(num),fit(5),err(5)
      real alpha(5,5),covar(5,5)
      integer flag,mfit
      real guess(2)

      data itmax,conv /50,1.0e-05/
      data chicon /2.5/


c  set the initial guesses

      flag = 0

      imax = 1
      do i=2,num
         if (flux(i) .gt. flux(imax)) imax = i
      end do

      if (guess(1).gt.0.) then
         if (imax .gt. num/2) then
            fit(1) = (flux(1) + flux(2) + flux(3)) / 3.0
         else
            fit(1) = (flux(num) + flux(num-1) + flux(num-2)) / 3.0
         end if
      else
         fit(1) = guess(1)
      end if

c      ip = imax + 1
c      im = imax - 1
c      t = (flux(im) - flux(imax)) / (wave(im) - wave(imax))
c      c = (t - (flux(im) - flux(ip))/(wave(im) - wave(ip))) /
c     $     (wave(im) - wave(ip))
c      b = t - c * (wave(im) + wave(imax))
c      a = flux(imax) - wave(imax) * ( b + wave(imax) * c)

c      fit(2) = 4.3 * ((a - 0.25 * b * b / c) - fit(1))
c      fit(3) = -0.5 * b / c

c     the 2.5 on the end is the root(2pi) which was missing

      fit(2) = 4.3 * (flux(imax) - fit(1)) * 2.5
      if (guess(2).gt.0.) then
         fit(3) = guess(2)
      else
         fit(3) = wave(imax)
      end if
      fit(4) = 0.05
      fit(5) = sl

      alamda = -1.0
      ochisq = 1000.0

      do iter=1,itmax
         oalam = alamda
         
         call mrqmin (wave,flux,sigma,num,fit,sg,mfit,
     $        covar,alpha,chisq,alamda,flag)      

         if (flag .ne. 0) then
            return
         endif

c         write (*,1000) iter,alamda,chisq,fit
 1000    format (i3,1x,1pe7.1,1x,e10.4,5(1x,0pf9.3))

         dchisq = (ochisq - chisq) / chisq
         ochisq = chisq
         rchisq = chisq / (num-mfit)
         if ((dchisq .lt. conv) .and. (oalam .gt. alamda)) go to 100
c         if ((rchisq .lt. chicon) .and. (oalam .gt. alamda)) go to 100
      end do

      iter = 0

 100  alamda = 0.0
      call mrqmin (wave,flux,sigma,num,fit,sg,mfit,
     $     covar,alpha,chisq,alamda,flag)      

      if (flag .ne. 0) then
         return
      endif

      ndof = num - mfit
      chisq = chisq / float(ndof)
      if (iter .eq. 0) chisq = - chisq

      do i=1,mfit
         err(i)=sqrt(covar(i,i))
      end do

      return
      end


c *******************************************************************
c
c  This subroutine evaluates a Voigt emission profile and its 
c  derivatives at a specified wavelength.  The parameters are:
c
c          a(1)  continuum level
c          a(2)  total line strength (area under Voigt)
c          a(3)  center wavelength
c          a(4)  Gaussian sigma component
c          a(5)  Lorentzian sigma
c
c  The full Gaussian sigma is the sum in quadrature of a(4) and
c  the instrumental gaussian sigma, si.
c
c *******************************************************************

      subroutine voigt (wave,a,si,vgt,dvda)

      parameter (root2=1.4142136e0)
      parameter (rootpi=1.7724539e0)

      real a(5),dvda(5)


      sgsq = a(4)**2 + si**2
      sg = sqrt(sgsq)
      r2sg = 1.0 / (root2 * sg)
      x = (wave - a(3)) * r2sg
      y = a(5) * r2sg


      rpsg = r2sg / rootpi

      call voi(x,y,v,dvdx,dvdy)

c     this fixes the intensity problem (the reported intensities were too high)
c
      a2sg = a(2) / sg
c  
c      a2sg = a(2) * rpsg

      vgt = a(1) + a2sg * v

      dvda(1) = 1.0
      dvda(2) = rpsg * v
      dvda(3) = - a2sg * r2sg * dvdx
      dvda(4) = - a2sg * a(4) * (v + x * dvdx + y * dvdy) / sgsq
      dvda(5) = a2sg * r2sg * dvdy

      return
      end

C***********************************************************************
C
C     This routine computes the normalized Voigt function and its partial 
C     derivatives.  The integral of this function is the value of the 
C     Gaussian sigma of the Voigt function.
C
C     The computation is valid for  y > 0, and the maximum relative error
C     for the Voigt function is < 2.E-6
C
C     Subroutine adapted from:
C     J. Humlicek, J. Quant. Spectrosc. Radiat. Transfer 21, 309 (1979)
C     by Philip L. Varghese        <820315.0112>
C
C     Modified to compute derivatives by Ted Williams 5/21/92
c     Modified to give correct value for intensity (?).
C
C***********************************************************************

      subroutine voi (x,y,v,dvdx,dvdy)


      real t(6), c(6), s(6)

      data t/0.314240376E+00, 0.947788391E+00, 0.159768264E+01,
     *       0.227950708E+01, 0.302063703E+01, 0.388972490E+01/
      data c/0.403621095E+00,-0.299993210E+00, 0.500980825E-02,
     *       0.399820281E-02,-0.956712138E-04, 0.199809468E-07/
      data s/0.555821146E+00, 0.922164680E-01,-0.619762811E-01,
     *       0.248076921E-02, 0.366661062E-04,-0.250346637E-06/

c          coefficients for non-normalized Voigt:
c      data c/0.101172805E01,-0.75197147E00,0.12557727E-01,
c     *       0.100220082E-01,-0.242068135E-03,0.500848061E-06/
c      data s/0.1393237E01,0.231152406E00,-0.155351466E00,
c     *       0.621836624E-02,0.919082986E-04,-0.627525958E-06/

      parameter (root2pi=2.50662827e0)


      v = 0.0
      dvdx = 0.0
      dvdy = 0.0

      y1 = y + 1.5
      y2 = y1 * y1

C  branch to region I or II depending on values of X and Y.

      if ((y.gt.0.85).or.(abs(x).lt.(18.1*y+1.65))) then

C  calculations for region I

         do i=1,6
            rm = x - t(i)
            dm = 1.0 / (rm * rm + y2)
            d1 = y1 * dm
            d2 = rm * dm
            d1d2 = d1 * d2
            rp = x + t(i)
            dp = 1.0 / (rp * rp + y2)
            d3 = y1 * dp
            d4 = rp * dp
            d3d4 = d3 * d4
            v  = v + c(i) * (d1 + d3) - s(i) * (d2 - d4)
            dvdx = dvdx - 2.0 *(c(i) * (d1d2 + d3d4)
     $             + s(i) * (0.5*(dm-dp) - d2*d2 + d4*d4))
            dvdy = dvdy + 2.0 * (s(i) * (d1d2 - d3d4)
     $             + c(i) * (0.5*(dm+dp) - d1*d1 - d3*d3))
         end do
      else

C  calculations for region II

         if (abs(x).lt.12.0) then
            v = exp(-x*x) / root2pi
            dvdx = -2.0 * x * v / root2pi
         end if

         y3 = y + 3.0
         do i=1,6
            rm = x - t(i)
            sm = rm * rm
            dm = 1.0 / (sm + y2)
            d1 = y1 * dm
            d2 = rm * dm
            tm = 1.0 / (sm + 2.25)
            d2tm = d2 * tm
            dmtm = dm * tm
            fm =  tm * (s(i)*y3*d2 + c(i)*(rm*d2 - 1.5*d1))
            rp = x + t(i)
            sp = rp * rp
            dp = 1.0 / (sp + y2)
            d3 = y1 * dp
            d4 = rp * dp
            tp = 1.0 / (sp + 2.25)
            d4tp = d4 * tp
            dptp = dp * tp
            fp = tp * (-s(i)*y3*d4 + c(i)*(rp*d4 - 1.5*d3))
            v = v + y * (fm + fp)
            dvdx = dvdx + y * (s(i)*y3*(dmtm - dptp)
     $           + 2.0 * (c(i)*(d2tm + d4tp)
     $           - fm*(d2 + rm*tm) - fp*(d4 + rp*tp)))
            dvdy = dvdy + fm + fp + y * (s(i)*(d2tm - d4tp)
     $           - 1.5*c(i)*(dmtm + dptp)
     $           - 2.0*(d1*fm + d3*fp))
         end do
      end if

      return
      end

c========================
c
c     Numerical Recipes subroutines for fitting.
c     

      subroutine mrqmin (x,y,sig,ndata,a,sg,mfit,covar,alpha,chisq,
     $                   alamda,flag)

      real x(ndata),y(ndata),sig(ndata),a(5),covar(5,5),alpha(5,5)
      real atry(5),beta(5),da(5)
      integer flag

      if (alamda .lt. 0.0) then
         alamda = 4.0
         call mrqcof (x,y,sig,ndata,a,sg,mfit,alpha,beta,chisq)
         ochisq = chisq
         do j=1,5
            atry(j) = a(j)
         end do
      end if

      do j=1,mfit
         do k=1,mfit
            covar(j,k) = alpha(j,k)
         end do
         covar(j,j) = alpha(j,j) * (1.0 + alamda)
         da(j) = beta(j)
      end do

      call gaussj_2 (covar,mfit,5,da,1,1,flag)

      if (flag .ne. 0) then
         return
      endif

      if (alamda .eq. 0.0) then
         return
      end if
      
      do j=1,mfit
         atry(j) = a(j) + da(j)
      end do

      call mrqcof (x,y,sig,ndata,atry,sg,mfit,covar,da,chisq)

      if (chisq .lt. ochisq) then
         alamda = 0.5 * alamda
         ochisq = chisq
         do j=1,mfit
            do k=1,mfit
               alpha(j,k) = covar(j,k)
            end do
            beta(j) = da(j)
            a(j) = atry(j)
         end do
      else
         alamda = 2.0 * alamda
         chisq = ochisq
      end if

      return
      end

c========================
c

      subroutine mrqcof (x,y,sig,ndata,a,sg,mfit,alpha,beta,chisq)

      real x(ndata),y(ndata),sig(ndata),alpha(5,5),beta(5),dvda(5)
      real a(5)

      do j=1,mfit
         do  k=1,j
            alpha(j,k) = 0.0
         end do
         beta(j) = 0.0
      end do
      chisq = 0.0

      do i=1,ndata
         call voigt (x(i),a,sg,v,dvda)
         sig2i = 1.0 / (sig(i) * sig(i))
         dy = y(i) - v
         do j=1,mfit
            wt = dvda(j) * sig2i
            do k=1,j
               alpha(j,k) = alpha(j,k) + wt * dvda(k)
            end do
            beta(j) = beta(j) + dy * wt
         end do
         chisq = chisq + dy * dy * sig2i
      end do

      do j=2,mfit
         do k=1,j-1
            alpha(k,j) = alpha(j,k)
         end do
      end do

      return
      end


c========================
c

      SUBROUTINE GAUSSJ_2(A,N,NP,B,M,MP,flag)
      PARAMETER (NMAX=50)
      integer flag
      DIMENSION A(NP,NP),B(NP,MP),IPIV(NMAX),INDXR(NMAX),INDXC(NMAX)
      DO 11 J=1,N
        IPIV(J)=0
11    CONTINUE
      DO 22 I=1,N
        BIG=0.
        DO 13 J=1,N
          IF(IPIV(J).NE.1)THEN
            DO 12 K=1,N
              IF (IPIV(K).EQ.0) THEN
                IF (ABS(A(J,K)).GE.BIG)THEN
                  BIG=ABS(A(J,K))
                  IROW=J
                  ICOL=K
                ENDIF
              ELSE IF (IPIV(K).GT.1) THEN
c                PAUSE 'Singular matrix'
                flag = 1
                return
              ENDIF
12          CONTINUE
          ENDIF
13      CONTINUE
        IPIV(ICOL)=IPIV(ICOL)+1
        IF (IROW.NE.ICOL) THEN
          DO 14 L=1,N
            DUM=A(IROW,L)
            A(IROW,L)=A(ICOL,L)
            A(ICOL,L)=DUM
14        CONTINUE
          DO 15 L=1,M
            DUM=B(IROW,L)
            B(IROW,L)=B(ICOL,L)
            B(ICOL,L)=DUM
15        CONTINUE
        ENDIF
        INDXR(I)=IROW
        INDXC(I)=ICOL
        IF (A(ICOL,ICOL).EQ.0.) then
c           PAUSE 'Singular matrix.'
           flag = 1
           return
        endif
        PIVINV=1./A(ICOL,ICOL)
        A(ICOL,ICOL)=1.
        DO 16 L=1,N
          A(ICOL,L)=A(ICOL,L)*PIVINV
16      CONTINUE
        DO 17 L=1,M
          B(ICOL,L)=B(ICOL,L)*PIVINV
17      CONTINUE
        DO 21 LL=1,N
          IF(LL.NE.ICOL)THEN
            DUM=A(LL,ICOL)
            A(LL,ICOL)=0.
            DO 18 L=1,N
              A(LL,L)=A(LL,L)-A(ICOL,L)*DUM
18          CONTINUE
            DO 19 L=1,M
              B(LL,L)=B(LL,L)-B(ICOL,L)*DUM
19          CONTINUE
          ENDIF
21      CONTINUE
22    CONTINUE
      DO 24 L=N,1,-1
        IF(INDXR(L).NE.INDXC(L))THEN
          DO 23 K=1,N
            DUM=A(K,INDXR(L))
            A(K,INDXR(L))=A(K,INDXC(L))
            A(K,INDXC(L))=DUM
23        CONTINUE
        ENDIF
24    CONTINUE
      RETURN
      END

