C-----------------------------------------------------------C
C      autocalibrate.f
C
C      Description
C
C      Allows a user to run the autosausage calibration routine
C      on a spectrum generated from a sausage, rather than on the
C      images of the sausage themselves (via a call to sausage)
C
C      The usage for this is:
C  autocalibrate (filter) (z_coarse) (sausage_file) [additional files]
C
C      Any number of sausage files may be calibrated at the same time,
C      Provided they have the same filter and z_coarse setting
C
C      Authors
C      Written by: Benjamin Shaya
C      In association with Mike McDonald and Sylvain Veilleux
C      For use with the Maryland-Magellan Tunable Filter (MMTF)
C
C      History:
C  6/2010 - Created 
C     (Ben, et al.,University of Maryland, College Park)
C
C-----------------------------------------------------------C

      program autocalibrate
        
        implicit none
        character(80) sausage
        character(80) filter,date,time,outfile,arg,getDate,getTime
        double precision a(4),b(4),n(4),rms
        integer z,nargs,i

C     Get command-line arguments.
        nargs = iargc()

        if (nargs .lt. 3) then
          print*,"Usage: "
          print*,"autocalibrate filter z_coarse" // 
     $     " sausage_file [additional sausages"
          stop
        end if     

        call getarg(1,filter)
        call getarg(2,arg)
        read(arg,*)z

C       This allows any number of sausages to be calibrated with a
C       single command.
C       Most users will probably not have a use for this feature.
        do i=3,nargs
          call getarg(i,sausage)
          write(outfile,'(a)')
     $     sausage(index(sausage,"/",.true.)+1:80)
          date = getDate()
          time = getTime()
          print*,sausage
          call findFit(sausage,filter,z,date,time,outFile)
        end do

      end program autocalibrate

C     Gets the date+time and formats it decently
C     This is not needed when using fits files, as fits files
C     contain date data.

      function getDate()
         character(80) getDate
         integer*4 today(3)
         call idate(today)   ! today(1)=day, (2)=month, (3)=year
         write (getDate,'(2(I2,"/"),I4)') 
     $    today(2),today(1),today(3)

      end function getDate

      function getTime()
         character(80) getTime
         integer*4 today(3), now(3)
         call itime(now)     ! now(1)=hour, (2)=minute, (3)=second
         write (getTime,'(I2,":",I2.2)') 
     $    now(1),now(2)

      end function getTime
