C
c     Description
c
c     Find the optical axis of the Fabry-Perot etalon, given centers of
c     point sources and their corresponding primary ghost reflections.
c     (The axis of reflection is the optical axis.)
c
c     Updated Mar 2010 - Now everything is automated. Using the custom
c                        MMTF slit mask, we know where the ghosts will
c                        be, roughly, and can bypass user input.
c
c     Authors
c
c     Ben Weiner, David Rupke, Michael McDonald
c
c     History
c
c     ???      BW    created
c     07jan15  DSNR  massaged outputs for clarity
c     10Mar    MM    Now fully automated
c
      program optaxis

      implicit none

      integer IFDEBUG,NUMCHIPS,MAXPTS
      parameter(IFDEBUG=0)
      parameter(NUMCHIPS=8)
      parameter(MAXPTS=1000)

      real xcorner11(NUMCHIPS),ycorner11(NUMCHIPS)
      real xcorner12(NUMCHIPS),ycorner12(NUMCHIPS)
      real xcorner21(NUMCHIPS),ycorner21(NUMCHIPS)
      real xcorner22(NUMCHIPS),ycorner22(NUMCHIPS)
      real xplot(5),yplot(5)
      real xp(MAXPTS),yp(MAXPTS),xg(MAXPTS),yg(MAXPTS)
      real vpx1,vpx2,vpy1,vpy2,pgrow,xgrow,vpxgrow,plg0,plg1,plg2,plg3
     $     ,plg4
      real xfp(MAXPTS),yfp(MAXPTS),xfg(MAXPTS),yfg(MAXPTS)
      real xfmid(MAXPTS),yfmid(MAXPTS)
      real xfmidmax,xfmidmin,yfmidmax,yfmidmin,xpmean,ypmean
      real xfsum,yfsum,xfrms,yfrms
      real xplot1,xplot2,yplot1,yplot2,xmarg,ymarg
      real xfmax,xfmin,yfmax,yfmin,siglev,sumsqsig

      integer i,j,ios,ilen,zoomplot,pgstat,psplot,img,ier
      integer nargs,npts,nchips,ncmean,nxpix,nypix,np(MAXPTS),ng(MAXPTS)

      character*80 arg,toplabel1,toplabel2,datfile,psfile
      character fname*120,clabel*2,infile*15

c     PGPLOT subroutine declaration.
      integer pgbeg

c     IMACS chip parameters.
      nchips = 8
      nxpix = 2048
      nypix = 4096

C     Get command-line arguments.
      nargs = iargc()
      if (nargs .lt. 1) then
         write (*,'(a,/,a,/,a,a)') "Usage: optaxis (input file)"
     $        ,"   Format of input file: ift####"
c     $        ,"   Chip(image) X(image) Y(image) "
c     $        ,"Chip(ghost) X(ghost) Y(ghost)"
         stop
      end if
      call getarg(1,arg)
      read (arg,*) fname

      ! ===================================================
      ! ==== Get centroid positions of stars and ghosts ===
      ! ===================================================
      npts=8
      do j=1,4
         np(j)=5
         ng(j)=3
      end do
      do j=5,8
         np(j)=8
         ng(j)=2
      end do
c     Load up images 2,3,5,8 and find star/ghost positions
      ! Image 5, has 4 stars in upper-left corner
      ! (277,3809), (793,3808), (328,3293), (843,3292)
      infile(1:7)=fname(1:7)
      infile(8:14)='c5.fits'
      call imopen(infile(1:14),1,img,ier)
      call getstar(img,277,3809,xp(1),yp(1))
      print *, 'Found star at ',xp(1),',',yp(1),' on chip 5'
      call getstar(img,793,3808,xp(2),yp(2))
      print *, 'Found star at ',xp(2),',',yp(2),' on chip 5'
      call getstar(img,328,3293,xp(3),yp(3))
      print *, 'Found star at ',xp(3),',',yp(3),' on chip 5'
      call getstar(img,843,3292,xp(4),yp(4))
      print *, 'Found star at ',xp(4),',',yp(4),' on chip 5'
      call imclos(img,ier)

      ! Image 8, has 4 stars in upper-right corner
      ! (1839,3808), (1325,3808), (1273,3292), (1789,3292)
      infile(8:14)='c8.fits'
      call imopen(infile(1:14),1,img,ier)
      call getstar(img,1839,3808,xp(5),yp(5))
      print *, 'Found star at ',xp(5),',',yp(5),' on chip 8'
      call getstar(img,1325,3808,xp(6),yp(6))
      print *, 'Found star at ',xp(6),',',yp(6),' on chip 8'
      call getstar(img,1273,3292,xp(7),yp(7))
      print *, 'Found star at ',xp(7),',',yp(7),' on chip 8'
      call getstar(img,1789,3292,xp(8),yp(8))
      print *, 'Found star at ',xp(8),',',yp(8),' on chip 8'
      call imclos(img,ier)

      ! Image 3, has 4 ghosts in upper-left corner
      ! (244,3868), (760,3868), (294,3354), (810,3352)
      infile(1:8)=fname(1:8)
      infile(8:14)='c3.fits'
      call imopen(infile(1:14),1,img,ier)
      call getstar(img,244,3868,xg(1),yg(1))
      print *, 'Found ghost at ',xg(1),',',yg(1),' on chip 3'
      call getstar(img,760,3868,xg(2),yg(2))
      print *, 'Found ghost at ',xg(2),',',yg(2),' on chip 3'
      call getstar(img,294,3354,xg(3),yg(3))
      print *, 'Found ghost at ',xg(3),',',yg(3),' on chip 3'
      call getstar(img,810,3352,xg(4),yg(4))
      print *, 'Found ghost at ',xg(4),',',yg(4),' on chip 3'
      call imclos(img,ier)

      ! Image 2, has 4 stars in upper-right corner
      ! (1809,3868), (1291,3868), (1241,3354), (1757,3352)
      infile(8:14)='c2.fits'
      call imopen(infile(1:14),1,img,ier)
      call getstar(img,1809,3868,xg(5),yg(5))
      print *, 'Found ghost at ',xg(5),',',yg(5),' on chip 2'
      call getstar(img,1291,3868,xg(6),yg(6))
      print *, 'Found ghost at ',xg(6),',',yg(6),' on chip 2'
      call getstar(img,1241,3354,xg(7),yg(7))
      print *, 'Found ghost at ',xg(7),',',yg(7),' on chip 2'
      call getstar(img,1757,3352,xg(8),yg(8))
      print *, 'Found ghost at ',xg(8),',',yg(8),' on chip 2'
      call imclos(img,ier)


!============================================================
!     Resume old operation here
      
c     First or second plot? 0 = first
      zoomplot=0
c     Have we written the postscript file?
      psplot=0

c     Compute focal plane locations of chip corners
      xfmin=1.e10
      xfmax=-1.e10
      yfmin=1.e10
      yfmax=-1.e10
      do j=1,nchips
         call chiptofp(j,1.,1.,xcorner11(j),ycorner11(j))
         call chiptofp(j,1.,real(nypix),xcorner12(j),ycorner12(j))
         call chiptofp(j,real(nxpix),1.,xcorner21(j),ycorner21(j))
         call chiptofp(j,real(nxpix),real(nypix),
     +        xcorner22(j),ycorner22(j))
         xfmin = min(xfmin,min(min(xcorner11(j),xcorner12(j)),
     +        min(xcorner21(j),xcorner22(j))))
         xfmax = max(xfmax,max(max(xcorner11(j),xcorner12(j)),
     +        max(xcorner21(j),xcorner22(j))))
         yfmin = min(yfmin,min(min(ycorner11(j),ycorner12(j)),
     +        min(ycorner21(j),ycorner22(j))))
         yfmax = max(yfmax,max(max(ycorner11(j),ycorner12(j)),
     +        max(ycorner21(j),ycorner22(j))))
         if (IFDEBUG .ge. 2) then
            write (*,'(4(1x,f8.2))') xcorner11(j),xcorner12(j),
     $           xcorner22(j),xcorner21(j)
            write (*,'(4(1x,f8.2))') ycorner11(j),ycorner12(j),
     $           ycorner22(j),ycorner21(j)
         end if
      end do

c     Open output file.
      ilen = index(fname,' ')-1
      datfile = fname(1:ilen)//'.dat'
      open (3,file=datfile,status='unknown')

c     Open plotter.
      pgstat = pgbeg(0,'/XS',1,1)
 175  call pgsch(1.1)
      call pgslw(2)
      call pgscf(2)
      call pgask (.true.)

c     Get plot limits
      xmarg = 0.05*(xfmax-xfmin)
      ymarg = 0.05*(yfmax-yfmin)
      xplot1 = xfmax+xmarg
      xplot2 = xfmin-xmarg
      yplot1 = yfmin-ymarg
      yplot2 = yfmax+ymarg
 200  continue
      if (zoomplot.eq.0) then
         call pgenv(xplot1,xplot2,yplot1,yplot2,1,0)
      else
         xplot1 = xfmidmin - (xfmidmax-xfmidmin)*0.05
         xplot2 = xfmidmax + (xfmidmax-xfmidmin)*0.05
         yplot1 = yfmidmin - (yfmidmax-yfmidmin)*0.05
         yplot2 = yfmidmax + (yfmidmax-yfmidmin)*0.05
         call pgenv(xplot1,xplot2,yplot1,yplot2,0,1)
      end if

c     Draw chip boundaries
      do j=1,nchips
         xplot(1) = xcorner11(j)
         yplot(1) = ycorner11(j)
         xplot(2) = xcorner12(j)
         yplot(2) = ycorner12(j)
         xplot(3) = xcorner22(j)
         yplot(3) = ycorner22(j)
         xplot(4) = xcorner21(j)
         yplot(4) = ycorner21(j)
         xplot(5) = xcorner11(j)
         yplot(5) = ycorner11(j)
         call pgsci(3)
         call pgline(5,xplot,yplot)
         write(clabel,'(i2)') j
         if (j.lt.5) then 
            call pgptxt(xplot(1)-200.,yplot(1)+200.,0.0,0.5,clabel)
         else
            call pgptxt(xplot(4)-200.,yplot(4)-400.,0.0,0.5,clabel)
         end if
         call pgsci(1)
      end do

      if (zoomplot.eq.1.AND.psplot.eq.0) then
         write(*,'(a)') "Individual points: [focal plane coordinates]"
         write(*,'(a3,7(a9))') "","Image","","Ghost","" ,"Midpoint",""
     $        ,"Outlier?"
         write(*,'(a3,7(a9))') "n","X(pix)","Y(pix)","X","Y","X","Y"
     $        ,"AD/MAD"
         write(3,'(/,a)') "Individual points: [focal plane coordinates]"
         write(3,'(a3,7(a9))') "","Image","","Ghost","" ,"Midpoint",""
     $        ,"Outlier?"
         write(3,'(a3,7(a9))') "n","X(pix)","Y(pix)","X","Y","X","Y"
     $        ,"N*MAD"
      end if
      call pgsch(2.)
c      xfsum = 0.0
c      yfsum = 0.0
c      xfsumsq = 0.0
c      yfsumsq = 0.0
      xfmidmax = -1e4
      xfmidmin =  1e4
      yfmidmax = -1e4
      yfmidmin =  1e4
      do i=1,npts
         call chiptofp(np(i),xp(i),yp(i),xfp(i),yfp(i))
         call chiptofp(ng(i),xg(i),yg(i),xfg(i),yfg(i))
         xfmid(i) = (xfp(i) + xfg(i))/2.0
         yfmid(i) = (yfp(i) + yfg(i))/2.0
         xfmidmax = max(xfmid(i),xfmidmax)
         xfmidmin = min(xfmid(i),xfmidmin)
         yfmidmax = max(yfmid(i),yfmidmax)
         yfmidmin = min(yfmid(i),yfmidmin)
         call pgpt1(xfp(i),yfp(i),17)
         call pgpt1(xfg(i),yfg(i),21)
         call pgpt1(xfmid(i),yfmid(i),5)
c         xfsum = xfsum + xfmid(i)
c         yfsum = yfsum + yfmid(i)
c         xfsumsq = xfsumsq + xfmid(i)**2
c         yfsumsq = yfsumsq + yfmid(i)**2         
         if (zoomplot.eq.1.AND.psplot.eq.0) then
            siglev = sqrt((xfsum-xfmid(i))**2+(yfsum-yfmid(i))**2)
     $           /sumsqsig
            write(*,'(i3,6(f9.2),f9.1)') i,xfp(i),yfp(i) ,xfg(i),yfg(i)
     $           ,xfmid(i),yfmid(i),siglev
            write(3,'(i3,6(f9.2),f9.1)') i,xfp(i),yfp(i) ,xfg(i),yfg(i)
     $           ,xfmid(i),yfmid(i),siglev
         end if
      end do
c     Used to do mean . . .
c      xfsum = xfsum / npts
c      yfsum = yfsum / npts
c      xfrms = sqrt(xfsumsq/npts - xfsum**2)
c      yfrms = sqrt(yfsumsq/npts - yfsum**2)
c     Now we do a median!
      call medmad(xfmid,npts,xfsum,xfrms)
      call medmad(yfmid,npts,yfsum,yfrms)
      sumsqsig = sqrt(xfrms**2+yfrms**2)
      call fptochip(xfsum,yfsum,ncmean,xpmean,ypmean)
c     
      call pgsci(2)
      call pgpt1(xfsum,yfsum,8)
      call pgsci(1)
      call pgsch(1.1)

c     Plot labels
      if (zoomplot.eq.0) then
         write(toplabel1,'(a,a10,a20,i3,2(a5,f8.2))')
     $        "Full field of view  ","","OA (CCD):  Chip",ncmean,"X="
     $        ,xpmean,"Y=",ypmean
      else 
         write(toplabel1,'(a,a7,a20,i3,2(a5,f8.2))')
     $        "Zoom into midpoints","","OA (CCD):  Chip",ncmean,"X="
     $        ,xpmean,"Y=",ypmean
      end if
      write(toplabel2,'(a,a27,a12,2(a5,f8.2))') "Input file = " ,fname
     $     ,"OA (FP):","X=",xfsum ,"Y=" ,yfsum
      call pglabel("X (focal plane, pix)","Y (focal plane, pix)",'')
      call pgmtxt('T',1.9,0.5,0.5,toplabel1)
      call pgmtxt('T',0.7,0.5,0.5,toplabel2)

c     Plot legend
      if (zoomplot.eq.0) then
         call pgqvp(0,vpx1,vpx2,vpy1,vpy2)
         pgrow = 0.06
         vpxgrow = pgrow*abs(vpx2-vpx1)
         xgrow = pgrow*abs(xplot2-xplot1)
         plg0 = 0.05
         plg1 = 0.15
         plg2 = 0.25
         plg3 = 0.35
         plg4 = 0.45
         call pgsvp(vpx1-vpxgrow,vpx2+vpxgrow,vpy1,vpy2)
         call pgswin(xplot1+xgrow,xplot2-xgrow,yplot1,yplot2)
         call pgsch(2.)
         call pgpt1(xplot2-xgrow,yplot2-plg1*abs(yplot1-yplot2),17)
         call pgpt1(xplot2-xgrow,yplot2-plg2*abs(yplot1-yplot2),21)
         call pgpt1(xplot2-xgrow,yplot2-plg3*abs(yplot1-yplot2),5)
         call pgsci(3)
         call pgpt1(xplot2-xgrow,yplot2-plg0*abs(yplot1-yplot2),19)
         call pgsci(2)
         call pgpt1(xplot2-xgrow,yplot2-plg4*abs(yplot1-yplot2),8)
         call pgsci(1)
         call pgsch(1.1)
         call pgmtxt('RV',1.5,1.-plg0,0.,'CCDs')
         call pgmtxt('RV',1.5,1.-plg1,0.,'Image')
         call pgmtxt('RV',1.5,1.-plg2,0.,'Ghost')
         call pgmtxt('RV',1.5,1.-plg3,0.,'Midpt')
         call pgmtxt('RV',1.5,1.-plg4,0.,'Opt.Ax.')
      end if

c     Print midpoint data.
      if (zoomplot.eq.1.AND.psplot.eq.0) then
         write(*,'(a)') "Midpoint:"
         write(*,'(a30,a5,a10,a10)') "","Chip" ,"X(pix)" ,"Y(pix)"
         write(*, '(a30,i5,f10.2,f10.2)') "Median [CCD coords.]",ncmean
     $        ,xpmean,ypmean
         write(*,'(a30,a5,f10.2,f10.2)') "Median [focal plane coords.]"
     $        ,"---",xfsum,yfsum
         write(*,'(a30,a5,f10.2,f10.2)') "Median abs. dev. [MAD]","---"
     $        ,xfrms,yfrms
         write(3,'(/,a)') "Midpoint:"
         write(3,'(a30,a5,a10,a10)') "","Chip" ,"X(pix)" ,"Y(pix)"
         write(3, '(a30,i5,f10.2,f10.2)') "Median [CCD coords.]",ncmean
     $        ,xpmean,ypmean
         write(3,'(a30,a5,f10.2,f10.2)') "Median [focal plane coords.]"
     $        ,"---",xfsum,yfsum
         write(3,'(a30,a5,f10.2,f10.2)') "Median abs. dev. [MAD]","---"
     $        ,xfrms,yfrms
c         if (xfrms .gt. 5. .or. yfrms .gt. 5.) then
c            write(*,*) "Warning: large scatter in the midpoint values:"
c            write(*,*)
c     $           "you probably have an outlier or misidentified ghost."
      end if

      if (zoomplot.eq.0) then
         zoomplot=1
         go to 200
      end if


      if (psplot.eq.0) then
c     Close output file
         ilen=index(datfile,' ')-1
         write(*,'(a,a30)') "Writing output data file: ",datfile
         close(3)
c     Close screen plot
         call pgend()
c     Check for and open postscript device
         ilen = index(fname,' ')-1
         psfile = fname(1:ilen)//'.ps'
         pgstat = pgbeg(0,psfile//'/ps',1,1)
         if (pgstat.ne.1) then
            write(*,*)
     $           "/PS device unavailable; not producing postscript."
         else
            write(*,'("Writing output plot: ",a30)') psfile
            zoomplot=0
            psplot = 1
            go to 175
         end if
      end if

c     Close postscript plot.
      call pgend()

      end

      SUBROUTINE GETSTAR(im,xin,yin,xout,yout)
      integer i,j,k,l,num,found
      integer im,xin,yin,xmax,ymax,ier,peakx(10),peaky(10)
      real*4 xout,yout,row(5000),zmax,line(7),flux(7)
      real*4 image(2048,4096),peakz(10),xcent,ycent
      

      ! Read FITS data into array
      do i=25,4075
         call imgl2r(im,row,i,ier)
         do j=25,2000
            image(j,i)=row(j)
         end do
      end do

      ! Go through region in 5x5 chunks, find peak flux in region
      zmax=0.
      do i=yin-200,yin+200,5   ! Loop over rows in region
         do j=xin-200,xin+200,5   ! Loop over columns in region
c            print *, j,i,image(j,i)

            do k=1,5
               peakz(k)=0.
               peakx(k)=0
               peaky(k)=0
            end do

            ! Find 6th brightest pixel in 5x5 region
            do k=i,i+4,1
               do l=j,j+4,1
                  num=1
                  found=0
                  do while (found.eq.0 .and. num.le.10)
                     if (image(l,k).gt.peakz(num)) then
                        found=1
                        peakz(num)=image(l,k)
                        peakx(num)=l
                        peaky(num)=k
                     end if
                     num=num+1
                  end do
               end do
            end do
            if (peakz(6).gt.zmax) then
               zmax=peakz(6)
               xmax=peakx(6)
               ymax=peaky(6)
            end if
         end do
      end do

      zmax=0.
      do i=xmax-3,xmax+3
         do j=ymax-3,ymax+3
            if (image(j,i).gt.zmax) then
               xmax=i
               ymax=j
               zmax=image(i,j)
            end if
         end do
      end do

      ! Make an array in x with length of 7, centered on x,y peak
      ! Find x centroid
      num=0
      do i=xmax-3,xmax+3
         num=num+1
         line(num)=i
         flux(num)=image(i,ymax)
      end do
      call centroid(line,flux,7,7,xmax,xcent)

      ! Make an array in y with length of 7, centered on x,y peak
      ! Find y centroid
      num=0
      do i=ymax-3,ymax+3
         num=num+1
         line(num)=i
         flux(num)=image(int(xcent),i)
      end do
      call centroid(line,flux,7,7,ymax,ycent)

      ! Repeat once more, using centroids instead of peaks
      num=0
      do i=int(xcent)-3,int(xcent)+3
         num=num+1
         line(num)=i
         flux(num)=image(i,int(ycent))
      end do
      call centroid(line,flux,7,7,int(xcent),xcent)

      num=0
      do i=int(ycent)-3,int(ycent)+3
         num=num+1
         line(num)=i
         flux(num)=image(int(xcent),i)
      end do
      call centroid(line,flux,7,7,int(ycent),ycent)

      xout=xcent
      yout=ycent
      RETURN
      END

C
C A Simple Centroid Algorithm
C
C A modification of the Modified Moment Method (Stone 1989,AJ,97,1227)
C that works well with small apertures.
C
C Copyleft (L) 1998 Kenneth J. Mighell (Kitt Peak National Observatory)
C
      SUBROUTINE CENTROID(X,Z,NN,N,X_IN,X_OUT)
      INTEGER          N, NN ! <-- assumes that 1 <= N <= NN
      REAL             X(NN), Z(NN), X_IN, X_OUT, BIG
      DOUBLE PRECISION DELTA, SUM1, SUM2, DIFF, XX
      DOUBLE PRECISION INTENSITY, POSITION, MINIMUM
      INTEGER          I, J , NITERATIONS
      PARAMETER        (BIG=1E30,NITERATIONS=10)
      MINIMUM = BIG
      DO I = 1,N
        INTENSITY = Z(I) ! <-- Z(I) is the intensity at the position X(I)
        MINIMUM = MIN(INTENSITY,MINIMUM)
      ENDDO
      XX = X_IN
      DELTA = 0D0
      DO J = 1,NITERATIONS
        XX = XX + DELTA
        SUM1 = 0D0
        SUM2 = 0D0
        DO I = 1,N
          POSITION = X(I)
          INTENSITY = Z(I)
          DIFF = MAX((INTENSITY-MINIMUM),0D0)
          SUM1 = SUM1 + (POSITION-XX)*DIFF
          SUM2 = SUM2 + DIFF
        ENDDO
        DELTA = SUM1/SUM2
      ENDDO
      X_OUT = XX
      RETURN
      END
